//
//  MutexTest.cpp
//
//  Created  by Setsu on 7/25/2010.
//  Modified by Setsu on May 30, 2013.
//  Copyright 2010-2013 RoundSquare Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
// 
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//

//----------------------------------------
//	include
//----------------------------------------
#include <Poco/Format.h>
#include <Poco/Runnable.h>
#include <Poco/Thread.h>
#include <Poco/Mutex.h>
#include <Poco/ScopedLock.h>
#include <Poco/Stopwatch.h>

#include <string>

#include "ScopedLogMessage.h"
#include "PrepareConsoleLogger.h"

//----------------------------------------
//	const
//----------------------------------------
const int kNumLoops = 1000;
const int kNumRunnable = 5;

//----------------------------------------
//	MutexRunnable
//----------------------------------------
template <class T>
class MutexRunnable: public Poco::Runnable
{
public:
	MutexRunnable(T& lock, volatile int& counter) :
		m_Lock(lock)
	,	m_Counter(counter)
	,	m_Ok(true)
	{
	}
	
	void run()
	{
		int lastCount = 0;
		for(int i=0; i<kNumLoops; ++i)
		{
			{	// scope for read lock
				Poco::ScopedLock<T> lock(m_Lock);

				lastCount = m_Counter;
				for(int k=0; k<100; ++k)
				{
					if(m_Counter != lastCount)
					{
						m_Ok = false;
					}
					Poco::Thread::yield();
				}
			}

			{	// scope for write lock
				Poco::ScopedLock<T> lock(m_Lock);

				for(int k=0; k<100; ++k)
				{
					--m_Counter;
					Poco::Thread::yield();
				}
				for(int k=0; k<100; ++k)
				{
					++m_Counter;
					Poco::Thread::yield();
				}
				++m_Counter;
				if(m_Counter <= lastCount)
				{
					m_Ok = false;
				}
			}
		}
	}
	
	bool ok() const
	{
		return m_Ok;
	}
	
private:
	T&				m_Lock;
	volatile int&	m_Counter;
	bool			m_Ok;
};

//----------------------------------------
//	TestMutex
//----------------------------------------
template <class T>
void TestMutex(ScopedLogMessage& msg, const std::string& name)
{
	msg.Message(Poco::format(" --- %s ---", name));

	Poco::Stopwatch stopwatch;
	stopwatch.start();

	T lock;
	int counter = 0;

	MutexRunnable<T>* pRunnable[kNumRunnable];
	Poco::Thread thread[kNumRunnable];

	for(int i=0; i<kNumRunnable; ++i)
	{
		pRunnable[i] = new MutexRunnable<T>(lock, counter);
		thread[i].start(*pRunnable[i]);
	}

	for(int i=0; i<kNumRunnable; ++i)
	{
		thread[i].join();
		if(!pRunnable[i]->ok())
		{
			msg.Message(Poco::format(" MutexRunnable[%i] failed!", i));
		}
		delete pRunnable[i];
	}

	stopwatch.stop();

	msg.Message(Poco::format("  counter = %i (expected: %i)", counter, kNumLoops*kNumRunnable));
	msg.Message(Poco::format("  Elapsed time = %.3fmSec", (1000.0 * stopwatch.elapsed()) / stopwatch.resolution()));
}

//----------------------------------------
//	main
//----------------------------------------
int main(int /*argc*/, char** /*argv*/)
{
	PrepareConsoleLogger logger(Poco::Logger::ROOT, Poco::Message::PRIO_INFORMATION);

	ScopedLogMessage msg("MutexTest ", "start", "end");

	TestMutex<Poco::Mutex>(msg, "Poco::Mutex test");
	TestMutex<Poco::FastMutex>(msg, "Poco::FastMutex test");

	return 0;
}
